async def test_simple_daemon_view(lona_app_context):
    """
    This test tests if a the state a daemonized view is hold by the server over
    page refreshes, until the view gets stopped by the user.
    """

    from playwright.async_api import async_playwright

    from lona.pytest import eventually

    def setup_app(app):
        from datetime import datetime

        from lona.html import Button, HTML, Div
        from lona import View

        # TODO: remove in 2.0
        @app.route('/async/')
        class AsyncDaemonView(View):
            def handle_request(self, request):
                self.daemonize()

                message = Div(f'view started: {datetime.now()}', _id='message')

                html = HTML(
                    message,
                    Button('Stop', _id='stop'),
                )

                self.show(html)
                self.await_input_event()
                message.set_text('view stopped')
                self.show(html)

        @app.route('/sync/')
        class SyncDaemonView(View):
            STOP_DAEMON_WHEN_VIEW_FINISHES = False

            def handle_stop_button_click(self, input_event):
                self.is_daemon = False
                self.message.set_text('view stopped')

            def handle_request(self, request):
                self.is_daemon = True

                self.message = Div(
                    f'view started: {datetime.now()}',
                    _id='message',
                )

                return HTML(
                    self.message,
                    Button(
                        'Stop',
                        _id='stop',
                        handle_click=self.handle_stop_button_click,
                    ),
                )

    context = await lona_app_context(setup_app)

    for url in ('/async/', '/sync/'):
        async with async_playwright() as p:
            browser = await p.chromium.launch()
            browser_context = await browser.new_context()
            page = await browser_context.new_page()

            # initial load
            await page.goto(context.make_url(url))
            await page.wait_for_selector('#message')
            message1 = await page.inner_text('#message')

            # second load
            await page.goto(context.make_url(url))
            await page.wait_for_selector('#message')
            message2 = await page.inner_text('#message')

            assert message1 == message2

            # stop view
            await page.click('#stop')

            for attempt in eventually():
                async with attempt:
                    message3 = await page.inner_text('#message')

                    assert message3 == 'view stopped'

            # check if views got removed from the server
            await page.goto('about:blank')

            for attempt in eventually():
                async with attempt:
                    view_runtime_count = len(
                        context.server._view_runtime_controller._view_runtimes)

                    assert view_runtime_count == 0


async def test_multi_tab_daemon_view(lona_app_context):
    """
    This test tests the HTML synchronization between two tabs that show same,
    daemonized view, using a TextInput
    """

    from playwright.async_api import async_playwright

    from lona.pytest import eventually

    def setup_app(app):
        from lona.html import TextInput, HTML
        from lona import View

        @app.route('/')
        class DaemonView(View):
            def handle_request(self, request):
                self.daemonize()

                html = HTML(
                    TextInput(value='initial', _id='input'),
                )

                self.show(html)
                self.await_input_event()

    context = await lona_app_context(setup_app)

    async def get_value(page):
        return await page.input_value('#input')

    async def set_value(page, value):
        return await page.fill('#input', str(value))

    async with async_playwright() as p:
        browser1 = await p.chromium.launch()
        browser_context1 = await browser1.new_context()
        page1 = await browser_context1.new_page()
        page2 = await browser_context1.new_page()

        browser2 = await p.chromium.launch()
        browser_context2 = await browser2.new_context()
        page3 = await browser_context2.new_page()

        # initial load
        await page1.goto(context.make_url('/'))
        await page2.goto(context.make_url('/'))
        await page3.goto(context.make_url('/'))

        assert (await get_value(page1)) == 'initial'
        assert (await get_value(page2)) == 'initial'
        assert (await get_value(page3)) == 'initial'

        # set value in first tab
        await set_value(page1, 'foo')

        for attempt in eventually():
            async with attempt:
                assert (await get_value(page1)) == 'foo'
                assert (await get_value(page2)) == 'foo'
                assert (await get_value(page3)) == 'initial'

        # set value in second tab
        await set_value(page2, 'bar')

        for attempt in eventually():
            async with attempt:
                assert (await get_value(page1)) == 'bar'
                assert (await get_value(page2)) == 'bar'
                assert (await get_value(page3)) == 'initial'



import pytest

from lona.routing import Router, Route


@pytest.mark.incremental()
class TestBasicRouting:
    def setup_method(self):
        self.routes = [
            Route('/foo/<arg>/', None),
            Route('/foo/', None),
            Route('/bar/<arg>/', None),
            Route('/', None),
        ]

        self.router = Router()
        self.router.add_routes(*self.routes)

    def test_resolve_without_arguments(self):
        match, route, match_info = self.router.resolve('/foo/')

        assert match
        assert route == self.routes[1]
        assert match_info == {}

    def test_resolve_with_argument(self):
        match, route, match_info = self.router.resolve('/foo/bar/')

        assert match
        assert route == self.routes[0]
        assert match_info == {'arg': 'bar'}

    def test_resolve_with_argument_2(self):
        match, route, match_info = self.router.resolve('/bar/foo/')

        assert match
        assert route == self.routes[2]
        assert match_info == {'arg': 'foo'}

    def test_resolve_root(self):
        match, route, match_info = self.router.resolve('/')

        assert match
        assert route == self.routes[3]
        assert match_info == {}

    def test_cant_resolve_unknown_route(self):
        match, route, match_info = self.router.resolve('/foobar/')

        assert not match


@pytest.mark.incremental()
class TestRoutesWithRegex:
    def setup_method(self):
        self.routes = [
            Route('/number/<number:[0-9]+>/', None),
        ]

        self.router = Router()
        self.router.add_routes(*self.routes)

    def test_match_regex(self):
        match, route, match_info = self.router.resolve('/number/14/')

        assert match
        assert route == self.routes[0]
        assert match_info == {'number': '14'}

    def test_doesnt_match_regex(self):
        match, route, match_info = self.router.resolve('/number/14a/')

        assert not match


@pytest.mark.incremental()
class TestOptionalTrailingSlash:
    def setup_method(self):
        self.routes = [
            Route('/foo(/)', None),
            Route('/bar/', None),
            Route('/foobar/<arg>(/)', None),
        ]

        self.router = Router()
        self.router.add_routes(*self.routes)

    def test_optional_slash_resolves_without_slash(self):
        match, route, match_info = self.router.resolve('/foo')

        assert match
        assert route == self.routes[0]
        assert match_info == {}

    def test_optional_slash_resolves_with_slash(self):
        match, route, match_info = self.router.resolve('/foo/')

        assert match
        assert route == self.routes[0]
        assert match_info == {}

    def test_optional_slash_resolves_without_slash_and_argument(self):
        match, route, match_info = self.router.resolve('/foobar/foobar')

        assert match
        assert route == self.routes[2]
        assert match_info == {'arg': 'foobar'}

    def test_optional_slash_resolves_with_slash_and_argument(self):
        match, route, match_info = self.router.resolve('/foobar/foobar/')

        assert match
        assert route == self.routes[2]
        assert match_info == {'arg': 'foobar'}

    def test_required_slash_doesnt_resolve_without_slash(self):
        match, route, match_info = self.router.resolve('/bar')

        assert not match

    def test_required_slash_resolves_with_slash(self):
        match, route, match_info = self.router.resolve('/bar/')

        assert match
        assert route == self.routes[1]
        assert match_info == {}


@pytest.mark.incremental()
class TestReverseMatching:
    def setup_method(self):
        routes = [
            Route('/foo/<arg>/', None, name='foo'),
            Route('/bar/', None, name='bar'),
            Route('/foobar/<arg>(/)', None, name='foobar'),
        ]
        self.router = Router()
        self.router.add_routes(*routes)

    def test_reverse_with_arg(self):
        url = self.router.reverse('foo', arg='bar')

        assert url == '/foo/bar/'

    def test_reverse_with_arg_and_optional_slash(self):
        url = self.router.reverse('foobar', arg='foobar')

        assert url == '/foobar/foobar'

    def test_reverse_without_arg(self):
        url = self.router.reverse('bar')

        assert url == '/bar/'

    def test_reverse_unknown_arg(self):
        with pytest.raises(ValueError, match='missing URL arg: arg'):
            self.router.reverse('foo', foo='bar')

    def test_reverse_unknown_name(self):
        with pytest.raises(ValueError, match="no route named 'baz' found"):
            self.router.reverse('baz')



import pytest


@pytest.mark.parametrize('rendering_setup', [
    'chromium:client-1',
    'chromium:client-2',
    'firefox:client-1',
    'firefox:client-2',
    'webkit:client-1',
    'webkit:client-2',
])
async def test_rendering(rendering_setup, lona_project_context):
    """
    This test tests all client side rendering features, using the rendering
    test view in the test project.

    The test has two phases: DOM tests, where DOM elements get manipulated, and
    a CSS phase, where the browsers style API gets used.

    The DOM phase is pretty simple and straightforward: The view creates and
    manipulates a simple DOM tree and sends it to the client to render it. The
    rendered HTML has to be equal to the servers representation of the DOM.

    In the CSS phase we can't just check the servers style attributes against
    the clients style attributes, because Lona does not define defaults for
    properties that are not set by the user. In the browser all CSS properties
    always have a value.

    TODO: remove in 2.0
    This test has a last phase, that tests node-list operations using the
    legacy widget API.
    """

    import json
    import os

    from playwright.async_api import async_playwright

    from lona.compat import get_client_version
    from lona.html import Widget, HTML

    TEST_PROJECT_PATH = os.path.join(__file__, '../../test_project')

    browser_name, client_version = rendering_setup.split(':')
    client_version = int(client_version[7:])

    context = await lona_project_context(
        project_root=TEST_PROJECT_PATH,
        settings=['settings.py'],
        settings_post_overrides={
            'CLIENT_VERSION': client_version,
        },
    )

    async def next_step(page, number):
        await page.locator('#lona #next-step').click()
        await page.wait_for_selector(f'#lona #step-label>#current:has-text("{number}")')

    async def get_computed_style(page):
        return await page.eval_on_selector(
            '#rendering-root > div',
            """
                element => {
                    const computedStyle = getComputedStyle(element);
                    const propertyNames = Array.from(computedStyle);
                    const values = {};

                    propertyNames.forEach(name => {
                        values[name] = computedStyle.getPropertyValue(name);
                    });

                    return values;
                }
            """,
        )

    def get_style():
        return context.server.state['rendering-root'][0].style

    async def check_default_styles(page):
        """
        All CSS properties have a specified default value.
        Because we use window.getComputedStyle() to determine CSS values,
        we are required to check our estimated defaults are correct.
        """

        computed_style = await get_computed_style(page)

        assert computed_style['top'] == 'auto'
        assert computed_style['right'] == 'auto'
        assert computed_style['bottom'] == 'auto'
        assert computed_style['left'] == 'auto'

    async def parse_json(page, locator):
        element = page.locator(locator)
        json_string = await element.inner_html()

        return json.loads(json_string)

    async with async_playwright() as p:
        browser = await getattr(p, browser_name).launch()
        browser_context = await browser.new_context()
        page = await browser_context.new_page()

        # start rendering test view
        await page.goto(context.make_url('/frontend/rendering/'))
        await page.wait_for_selector('#lona h2:has-text("Rendering Test")')

        rendering_root_element = page.locator('#lona #rendering-root')

        # DOM tests ###########################################################
        for step in range(1, 22):
            await next_step(page, step)

            # get rendered html
            html_string = await rendering_root_element.inner_html()

            # parse and compare html
            html = HTML(f'<div id="rendering-root">{html_string}</div>')

            if isinstance(html, Widget):
                html = html[0]

            assert html.nodes == context.server.state['rendering-root'].nodes

        # CSS tests ###########################################################
        await check_default_styles(page)

        # 22 Set Style
        await next_step(page, 22)

        computed_style = await get_computed_style(page)

        assert computed_style['top'] == '1px'
        assert computed_style['right'] == '2px'
        assert computed_style['bottom'] == 'auto'
        assert computed_style['left'] == 'auto'

        # 23 Add Style
        await next_step(page, 23)

        computed_style = await get_computed_style(page)
        style = get_style()

        assert computed_style['top'] == '1px'
        assert computed_style['right'] == '2px'
        assert computed_style['bottom'] == '3px'
        assert computed_style['left'] == 'auto'

        # 24 Remove Style
        await next_step(page, 24)

        computed_style = await get_computed_style(page)
        style = get_style()

        assert 'top' not in style

        assert computed_style['top'] == 'auto'
        assert computed_style['right'] == '2px'
        assert computed_style['bottom'] == '3px'
        assert computed_style['left'] == 'auto'

        # 25 Reset Style
        await next_step(page, 25)

        computed_style = await get_computed_style(page)
        style = get_style()

        assert 'right' not in style
        assert 'bottom' not in style

        assert computed_style['top'] == 'auto'
        assert computed_style['right'] == 'auto'
        assert computed_style['bottom'] == 'auto'
        assert computed_style['left'] == '4px'

        # 26 Clear Style
        await next_step(page, 26)
        await check_default_styles(page)

        # widget data tests ###################################################
        for step in range(27, 39):
            await next_step(page, step)

            server_widget_data = await parse_json(
                page,
                '#lona #server-widget-data',
            )

            client_widget_data = await parse_json(
                page,
                '#lona #client-widget-data',
            )

            assert server_widget_data == client_widget_data

        # legacy widgets tests ################################################
        # TODO: remove in 2.0

        if get_client_version() != 1:
            return

        for step in range(39, 45):
            await next_step(page, step)

            client_html_string = await rendering_root_element.inner_html()

            client_html = HTML(client_html_string)
            server_html = HTML(str(context.server.state['rendering-root']))[0]

            assert client_html.nodes == server_html.nodes



def setup_app(app):
    from datetime import datetime

    from lona import View

    @app.route('/')
    class HoweView(View):
        def handle_request(self, request):
            return 'HOME'

    @app.route('/daemonized-view/')
    class DaemonizedView(View):
        def handle_request(self, request):
            self.daemonize()
            self.set_title('Daemonized View')
            self.show(f'View started: {datetime.now()}')

            self.await_input_event()


async def test_title(lona_app_context):
    """
    This test tests page titles on non-daemon views and daemon views
    """

    from playwright.async_api import async_playwright

    context = await lona_app_context(setup_app)

    async with async_playwright() as p:
        browser = await p.chromium.launch()
        browser_context = await browser.new_context()
        page = await browser_context.new_page()

        # test default title
        await page.goto(context.make_url('/'))
        await page.wait_for_selector('#lona:has-text("HOME")')
        assert await page.title() == 'Lona'

        # get timestamp of the daemonized view
        await page.goto(context.make_url('/daemonized-view/'))
        await page.wait_for_selector('#lona:has-text("View started: ")')
        timestamp = await page.inner_text('#lona')

        # test titles
        assert await page.title() == 'Daemonized View'

        # reload tab
        await page.reload()
        await page.wait_for_selector('#lona:has-text("View started: ")')

        assert await page.inner_text('#lona') == timestamp
        assert await page.title() == 'Daemonized View'



import pytest


async def get_frontend_uptime(page):
    return (await page.locator('#application-started').all_text_contents())[0]


async def get_exception_info(page):
    info = {
        'type': '',
        'args': '',
    }

    await page.wait_for_selector('#lona h2:has-text("500")')

    info['type'] = (
        await page.locator('#exception-type').all_text_contents())[0]

    info['args'] = (
        await page.locator('#exception-args').all_text_contents())[0]

    return info


@pytest.mark.parametrize('url_prefix', ['dict-', ''])
@pytest.mark.parametrize('client_version', [1, 2])
@pytest.mark.parametrize('browser_name', ['chromium', 'firefox', 'webkit'])
async def test_interactive_responses(
        browser_name,
        client_version,
        url_prefix,
        lona_project_context,
):

    import os

    from playwright.async_api import async_playwright, expect

    from lona.pytest import eventually

    context = await lona_project_context(
        project_root=os.path.join(__file__, '../../test_project'),
        settings=['settings.py'],
        settings_post_overrides={
            'CLIENT_VERSION': client_version,
        },
    )

    async with async_playwright() as p:
        browser = await getattr(p, browser_name).launch()
        browser_context = await browser.new_context()
        page = await browser_context.new_page()

        # empty response
        await page.goto(context.make_url(f'/{url_prefix}responses/interactive/'))
        await page.click('#lona a#empty-response')
        await expect(page.locator('#lona')).to_be_empty()

        # node response
        await page.goto(context.make_url(f'/{url_prefix}responses/interactive/'))
        await page.click('#lona a#node-response')
        await page.wait_for_selector('#lona p:has-text("Node Response")')

        # string response
        await page.goto(context.make_url(f'/{url_prefix}responses/interactive/'))
        await page.click('#lona a#string-response')
        await page.wait_for_selector('#lona:has-text("String Response")')

        # template response
        await page.goto(context.make_url(f'/{url_prefix}responses/interactive/'))
        await page.click('#lona a#template-response')
        await page.wait_for_selector('#lona h2:has-text("Template Response")')

        # template string response
        await page.goto(context.make_url(f'/{url_prefix}responses/interactive/'))
        await page.click('#lona a#template-string-response')
        await page.wait_for_selector('#lona:has-text("Template String Response")')

        # redirect response
        url = f'/{url_prefix}responses/interactive/'

        await page.goto(context.make_url(url))
        await page.wait_for_url(url)

        frontend_uptime = await get_frontend_uptime(page)

        await page.click('#lona a#redirect-response')
        await page.wait_for_url('/')

        assert (await get_frontend_uptime(page)) == frontend_uptime

        # http redirect response
        url = f'/{url_prefix}responses/interactive/'

        await page.goto(context.make_url(url))
        await page.wait_for_url(url)

        frontend_uptime = await get_frontend_uptime(page)

        await page.click('#lona a#http-redirect-response')
        await page.wait_for_url('/')

        for attempt in eventually():
            async with attempt:
                assert (await get_frontend_uptime(page)) != frontend_uptime

        # file response
        await page.goto(context.make_url(f'/{url_prefix}responses/interactive/'))
        await page.click('#lona a#file-response')

        exception_info = await get_exception_info(page)

        assert exception_info['type'] == 'RuntimeError'
        assert exception_info['args'] == "('JSON, binary and file responses and headers are only available in non-interactive mode',)"

        # json response
        await page.goto(context.make_url(f'/{url_prefix}responses/interactive/'))
        await page.click('#lona a#json-response')

        exception_info = await get_exception_info(page)

        assert exception_info['type'] == 'RuntimeError'
        assert exception_info['args'] == "('JSON, binary and file responses and headers are only available in non-interactive mode',)"

        # binary response
        await page.goto(context.make_url(f'/{url_prefix}responses/interactive/'))
        await page.click('#lona a#binary-response')

        exception_info = await get_exception_info(page)

        assert exception_info['type'] == 'RuntimeError'
        assert exception_info['args'] == "('JSON, binary and file responses and headers are only available in non-interactive mode',)"

        # custom headers response
        await page.goto(context.make_url(f'/{url_prefix}responses/interactive/'))
        await page.click('#lona a#custom-headers-response')

        exception_info = await get_exception_info(page)

        assert exception_info['type'] == 'RuntimeError'
        assert exception_info['args'] == "('JSON, binary and file responses and headers are only available in non-interactive mode',)"


@pytest.mark.parametrize('url_prefix', ['dict-', ''])
async def test_non_interactive_responses(url_prefix, lona_project_context):
    import os

    from aiohttp import ClientSession

    context = await lona_project_context(
        project_root=os.path.join(__file__, '../../test_project'),
        settings=['settings.py'],
    )

    async with ClientSession() as session:

        # empty response
        url = context.make_url(f'/{url_prefix}responses/non-interactive/empty-response')

        async with session.get(url) as response:
            assert response.status == 200
            assert (await response.text()) == ''

        # node response
        url = context.make_url(f'/{url_prefix}responses/non-interactive/node-response')

        async with session.get(url) as response:
            assert response.status == 200

            text = await response.text()

            assert '<p' in text
            assert 'Node Response' in text

        # string response
        url = context.make_url(f'/{url_prefix}responses/non-interactive/string-response')

        async with session.get(url) as response:
            assert response.status == 200
            assert (await response.text()) == 'String Response'

        # template response
        url = context.make_url(f'/{url_prefix}responses/non-interactive/template-response')

        async with session.get(url) as response:
            assert response.status == 200
            assert '<h2>Template Response</h2>' in (await response.text())

        # template string response
        url = context.make_url(f'/{url_prefix}responses/non-interactive/template-string-response')

        async with session.get(url) as response:
            assert response.status == 200
            assert (await response.text()) == 'Template String Response'

        # redirect response
        url = context.make_url(f'/{url_prefix}responses/non-interactive/redirect-response')

        async with session.get(url, allow_redirects=False) as response:
            assert response.status == 302
            assert response.headers['Location'] == '/'

        # http redirect response
        url = context.make_url(f'/{url_prefix}responses/non-interactive/http-redirect-response')

        async with session.get(url, allow_redirects=False) as response:
            assert response.status == 302
            assert response.headers['Location'] == '/'

        # file response
        path = os.path.join(
            os.path.dirname(__file__),
            '../doc/content/logo.svg',
        )

        url = context.make_url(f'/{url_prefix}responses/non-interactive/file-response')

        async with session.get(url) as response:
            assert response.status == 200
            assert (await response.content.read()) == open(path, 'rb').read()

        # json response
        url = context.make_url(f'/{url_prefix}responses/non-interactive/json-response')

        async with session.get(url) as response:
            assert response.status == 200
            assert response.content_type == 'application/json'
            assert (await response.json()) == {'foo': 'bar'}

        # binary response
        path = os.path.join(
            os.path.dirname(__file__),
            '../doc/content/logo.svg',
        )

        url = context.make_url(f'/{url_prefix}responses/non-interactive/binary-response')

        async with session.get(url) as response:
            assert response.status == 200
            assert response.content_type == 'image/svg+xml'
            assert (await response.content.read()) == open(path, 'rb').read()

        # custom headers response
        url = context.make_url(f'/{url_prefix}responses/non-interactive/custom-headers-response')

        async with session.get(url) as response:
            assert response.status == 418
            assert response.headers['HEADER-1'] == 'foo'
            assert response.headers['HEADER-2'] == 'bar'



import pytest


@pytest.mark.dependency(depends=[
    # see https://github.com/RKrahl/pytest-dependency/issues/55
    'tests/test_0001_html.py::TestAttributeDict::test_get_initial_values',
    'tests/test_0001_html.py::TestAttributeDict::test_set_attributes',
    'tests/test_0001_html.py::TestAttributeDict::test_value_cant_be_dict',
    'tests/test_0001_html.py::TestAttributeDict::test_cant_use_id_key',
    'tests/test_0001_html.py::TestAttributeDict::test_empty_dict_is_false',
    'tests/test_0001_html.py::TestAttributeDict::test_non_empty_dict_is_true',
    'tests/test_0001_html.py::TestAttributeDict::test_pop_existing_key',
    'tests/test_0001_html.py::TestAttributeDict::test_pop_unknown_key',
    'tests/test_0001_html.py::TestAttributeDict::test_pop_unknown_key_with_default',
    'tests/test_0001_html.py::TestAttributeDict::test_pop_existing_key_with_default',
    'tests/test_0001_html.py::TestAttributeDict::test_pop_expects_at_most_2_arguments',
    'tests/test_0001_html.py::TestAttributeDict::test_del_existing_key',
    'tests/test_0001_html.py::TestAttributeDict::test_del_unknown_key',
    'tests/test_0001_html.py::TestAttributeList::test_add_existing_element_does_nothing',
    'tests/test_0001_html.py::TestAttributeList::test_add_one_element',
    'tests/test_0001_html.py::TestAttributeList::test_cant_add_dict',
    'tests/test_0001_html.py::TestAttributeList::test_cant_set_dict',
    'tests/test_0001_html.py::TestAttributeList::test_cant_set_list_with_dict',
    'tests/test_0001_html.py::TestAttributeList::test_clear',
    'tests/test_0001_html.py::TestAttributeList::test_clear_empty_list',
    'tests/test_0001_html.py::TestAttributeList::test_default_class_list_is_empty',
    'tests/test_0001_html.py::TestAttributeList::test_default_id_list_is_empty',
    'tests/test_0001_html.py::TestAttributeList::test_default_list_has_zero_len',
    'tests/test_0001_html.py::TestAttributeList::test_empty_list_is_false',
    'tests/test_0001_html.py::TestAttributeList::test_equality_ignored_duplicates',
    'tests/test_0001_html.py::TestAttributeList::test_equality_ignores_order',
    'tests/test_0001_html.py::TestAttributeList::test_extend',
    'tests/test_0001_html.py::TestAttributeList::test_extend_ignores_duplicates',
    'tests/test_0001_html.py::TestAttributeList::test_in_keyword',
    'tests/test_0001_html.py::TestAttributeList::test_initial_class_can_be_list',
    'tests/test_0001_html.py::TestAttributeList::test_initial_class_can_be_passed_via_kwargs',
    'tests/test_0001_html.py::TestAttributeList::test_initial_class_can_be_space_separated_str',
    'tests/test_0001_html.py::TestAttributeList::test_initial_id_can_be_list',
    'tests/test_0001_html.py::TestAttributeList::test_initial_id_can_be_passed_via_kwargs',
    'tests/test_0001_html.py::TestAttributeList::test_initial_id_can_be_space_separated_str',
    'tests/test_0001_html.py::TestAttributeList::test_initial_id_cant_be_dict',
    'tests/test_0001_html.py::TestAttributeList::test_len_returns_number_of_elements',
    'tests/test_0001_html.py::TestAttributeList::test_non_empty_list_is_true',
    'tests/test_0001_html.py::TestAttributeList::test_non_equality',
    'tests/test_0001_html.py::TestAttributeList::test_remove_existing_element',
    'tests/test_0001_html.py::TestAttributeList::test_remove_unknown_element',
    'tests/test_0001_html.py::TestAttributeList::test_set_class_list',
    'tests/test_0001_html.py::TestAttributeList::test_set_id_list',
    'tests/test_0001_html.py::TestAttributeList::test_toggle_existing_element',
    'tests/test_0001_html.py::TestAttributeList::test_toggle_unknown_element',
], scope='session')
def test_node_api():
    pass



def setup_app(app):
    from lona.html import TextInput, FOCUS
    from lona import View

    @app.route('/')
    class MyView(View):
        def handle_request(self, request):
            self.show(TextInput(events=[FOCUS]))
            self.await_focus()
            self.show('SUCCESS')


async def test_focus_events(lona_app_context):
    from playwright.async_api import async_playwright

    context = await lona_app_context(setup_app)

    async with async_playwright() as p:
        browser = await p.chromium.launch()
        browser_context = await browser.new_context()
        page = await browser_context.new_page()

        await page.goto(context.make_url())
        assert await page.inner_text('#lona') != 'SUCCESS'
        await page.focus('input')
        await page.wait_for_selector('#lona:has-text("SUCCESS")')



from __future__ import annotations

from typing import Union, Dict, Any
from dataclasses import dataclass

from lona.html.abstract_node import AbstractNode
from lona.events.input_event import InputEvent


@dataclass
class AbstractResponse:
    @property
    def interactive(self) -> bool:
        raise NotImplementedError()


@dataclass
class Response(AbstractResponse):
    body: Any = None
    status: int | None = None
    content_type: str | None = None
    headers: Dict[str, str] | None = None
    text: str | None = None
    charset: str | None = None

    def __post_init__(self):
        if self.body and self.text:
            raise ValueError('body and text are not allowed together')

    @property
    def interactive(self) -> bool:
        if (self.body or self.status or self.content_type or
                self.headers or self.charset):

            return False

        return True


@dataclass
class HtmlResponse(AbstractResponse):
    html: str | AbstractNode
    status: int | None = None
    content_type: str | None = None
    headers: dict | None = None

    @property
    def interactive(self) -> bool:
        if self.status or self.content_type or self.headers:
            return False

        return True


@dataclass
class TemplateResponse(AbstractResponse):
    name: str
    context: dict | None
    status: int | None = None
    content_type: str | None = None
    headers: dict | None = None

    @property
    def interactive(self) -> bool:
        if self.status or self.content_type or self.headers:
            return False

        return True


@dataclass
class TemplateStringResponse(AbstractResponse):
    string: str
    context: dict | None = None
    status: int | None = None
    content_type: str | None = None
    headers: dict | None = None

    @property
    def interactive(self) -> bool:
        if self.status or self.content_type or self.headers:
            return False

        return True


@dataclass
class JsonResponse(AbstractResponse):
    data: Any
    status: int | None = None
    content_type: str | None = None
    headers: dict | None = None

    @property
    def interactive(self) -> bool:
        return False


@dataclass
class FileResponse(AbstractResponse):
    path: str
    status: int | None = None
    headers: dict | None = None

    @property
    def interactive(self) -> bool:
        return False


@dataclass
class RedirectResponse(AbstractResponse):
    url: str
    headers: dict | None = None

    @property
    def interactive(self) -> bool:
        if self.headers:
            return False

        return True


@dataclass
class HttpRedirectResponse(AbstractResponse):
    url: str
    headers: dict | None = None

    @property
    def interactive(self) -> bool:
        if self.headers:
            return False

        return True


# response parsing ############################################################
# view return values
def _parse_legacy_view_return_value(
        return_value: dict,
) -> AbstractResponse:

    key_words = {
        'text',
        'body',
        'redirect',
        'http_redirect',
        'template',
        'template_string',
        'file',
        'json',
    }

    # check keys
    if len(set(return_value) & key_words) != 1:
        raise RuntimeError(
            'response dicts have to contain exactly one keyword of: {}'.format(  # NOQA: FS002
                ', '.join(key_words),
            ),
        )

    # redirects
    if 'redirect' in return_value:
        return RedirectResponse(
            url=str(return_value['redirect']),
            headers=return_value.get('headers', None),
        )

    # http redirect
    if 'http_redirect' in return_value:
        return HttpRedirectResponse(
            url=str(return_value['http_redirect']),
            headers=return_value.get('headers', None),
        )

    # json
    if 'json' in return_value:
        return JsonResponse(
            data=return_value['json'],
            status=return_value.get('status', None),
            content_type=return_value.get('content_type', None),
            headers=return_value.get('headers', None),
        )

    # file
    if 'file' in return_value:
        return FileResponse(
            path=return_value['file'],
            status=return_value.get('status', None),
            headers=return_value.get('headers', None),
        )

    context: dict = return_value

    if 'context' in context:
        context = context['context']

    # template
    if 'template' in return_value:
        return TemplateResponse(
            name=str(return_value['template']),
            context=context,
            status=return_value.get('status', None),
            content_type=return_value.get('content_type', None),
            headers=return_value.get('headers', None),
        )

    # template string
    if 'template_string' in return_value:
        return TemplateStringResponse(
            string=str(return_value['template_string']),
            context=context,
            status=return_value.get('status', None),
            content_type=return_value.get('content_type', None),
            headers=return_value.get('headers', None),
        )

    return Response(
        body=return_value.get('body', None),
        status=return_value.get('status', None),
        content_type=return_value.get('content_type', None),
        headers=return_value.get('headers', None),
        text=return_value.get('text', None),
        charset=return_value.get('charset', None),
    )


def parse_view_return_value(
        return_value: AbstractResponse | AbstractNode | str | None | dict,
        interactive: bool,
) -> AbstractResponse | None:

    if return_value is None:
        return return_value

    # legacy view return values
    # TODO: remove in 2.0
    if isinstance(return_value, dict):
        return_value = _parse_legacy_view_return_value(
            return_value=return_value,
        )

    if isinstance(return_value, (str, AbstractNode)):
        return_value = HtmlResponse(html=return_value)

    # check for non-interactive features
    if interactive:

        # some response types can not be used in interactive views
        if isinstance(return_value, (JsonResponse, FileResponse)):
            raise RuntimeError(
                'JSON, binary and file responses and headers are only available in non-interactive mode',
            )

        # some response features can not be used in interactive views
        if (not isinstance(return_value, (RedirectResponse,
                                          HttpRedirectResponse)) and
                not return_value.interactive):

            raise RuntimeError(
                'JSON, binary and file responses and headers are only available in non-interactive mode',
            )

    if isinstance(return_value, AbstractResponse):
        return return_value

    raise RuntimeError(
        f'expected string, Lona Node or Lona response object. Got {type(return_value)}',
    )


# input event handler return values
def _parse_legacy_input_event_handler_return_value(
        return_value: Dict[str, str],
) -> RedirectResponse | HttpRedirectResponse:

    # TODO: remove in 2.0
    supported_keys = ('redirect', 'http_redirect')

    if (len(return_value.keys()) > 1 or
            list(return_value.keys())[0] not in supported_keys):

        raise ValueError(f'response dict has unexpected keys ({repr(list(return_value.keys()))})')

    redirect_type, redirect_target = list(return_value.items())[0]

    if not isinstance(redirect_target, str):
        raise ValueError('redirect target has to be string')

    if redirect_type == 'http_redirect':
        return HttpRedirectResponse(url=redirect_target)

    return RedirectResponse(url=redirect_target)


def parse_input_event_handler_return_value(
        return_value: Union[InputEvent, RedirectResponse, HttpRedirectResponse,
                            None, Dict[str, str]],
) -> InputEvent | RedirectResponse | HttpRedirectResponse | None:

    if return_value is None:
        return return_value

    # legacy input event handler return values
    # TODO: remove in 2.0
    if isinstance(return_value, dict):
        return_value = _parse_legacy_input_event_handler_return_value(
            return_value=return_value,
        )

    if isinstance(return_value, (InputEvent, RedirectResponse,
                                 HttpRedirectResponse)):

        return return_value

    raise RuntimeError(
        f'expected RedirectResponse, HttpRedirectResponse, InputEvent, dict or NoneType. Got {type(return_value)}',
    )



# TODO: remove LonaView and LonaApp in 2.0

from .exceptions import *  # NOQA: F403
from .responses import *  # NOQA: F403
from .routing import MATCH_ALL, Route
from .errors import *  # NOQA: F403
from .view import View as LonaView
from .app import App as LonaApp
from .view import View
from .app import App

VERSION = (1, 12, 3)
VERSION_STRING = '.'.join(str(i) for i in VERSION)



from lona import View


def setup_app(app):
    app.settings.TEMPLATE_EXTRA_FILTERS = {
        'custom_reverse': lambda string: string[::-1],
    }

    app.add_template('template_filters.html', "{{'foo'|custom_reverse}}")

    @app.route('/')
    class TemplateFiltersView(View):
        def handle_request(self, request):
            return {
                'template': 'template_filters.html',
            }


async def test_template_filters(lona_app_context):
    from playwright.async_api import async_playwright

    context = await lona_app_context(setup_app)

    async with async_playwright() as playwright:
        browser = await playwright.chromium.launch()
        browser_context = await browser.new_context()
        page = await browser_context.new_page()

        await page.goto(context.make_url('/'))
        await page.wait_for_selector('#lona:has-text("oof")')



from lona.html.data_binding.select2 import Select2, Option2
from lona.html.data_binding.inputs import *  # NOQA: F403
from lona.html.data_binding.select import Select, Option
from lona.html.parsing import _setup_node_classes_cache
from lona.events.event_types import *  # NOQA: F403
from lona.html.widgets import HTML as HTML1
from lona.html.parsing import HTML as HTML2
from lona.html.nodes import *  # NOQA: F403
from lona.compat import get_client_version
from lona.html.widget import Widget

_setup_node_classes_cache()


def HTML(*args, **kwargs):
    if get_client_version() == 1:
        return HTML1(*args, **kwargs)

    return HTML2(*args, **kwargs)



from __future__ import annotations

from tempfile import TemporaryDirectory
from argparse import Namespace
from typing import List
import os

TEST_PROJECT_PATH = os.path.join(
    os.path.dirname(__file__),
    '../test_project',
)


def _generate_command_line_args(
        destination: str,
        clean: bool = False,
        static_dirs: List[str] | None = None,
) -> Namespace:

    namespace = Namespace(

        # basic configuration
        project_root=TEST_PROJECT_PATH,
        settings=['settings.py'],
        settings_pre_overrides={},
        settings_post_overrides={},
        debug_mode='',
        log_level='info',
        loggers=[],
        syslog_priorities='auto',

        # collect-static specific arguments
        destination=destination,
        clean=clean,
        silent=False,
        dry_run=False,
    )

    if static_dirs:
        namespace.settings_post_overrides['STATIC_DIRS'] = static_dirs

    return namespace


def _check_javascript_client_src_files(destination: str) -> None:
    import lona

    CLIENT_ROOT = os.path.join(
        os.path.dirname(lona.__file__),
        'client/_lona/client/',
    )

    for client_src_file in os.listdir(CLIENT_ROOT):
        abs_path = os.path.join(destination, '_lona/client/', client_src_file)

        # check existence
        assert os.path.exists(abs_path)

        # check content
        original = open(os.path.join(CLIENT_ROOT, client_src_file), 'r').read()
        copy = open(abs_path, 'r').read()

        assert copy == original


def _check_test_project_static_files(destination: str):

    # files in test_project/static
    static_root = os.path.join(TEST_PROJECT_PATH, 'static')

    for static_file in os.listdir(static_root):
        abs_path = os.path.join(destination, static_file)

        # check existence
        assert os.path.exists(abs_path)

        # check content
        original = open(os.path.join(static_root, static_file), 'r').read()
        copy = open(abs_path, 'r').read()

        assert copy == original

    # files defined in View.STATIC_FILES and AbstractNode.STATIC_FILES
    static_files = [

        # contains: (original_path, copy_path, )
        (os.path.join(TEST_PROJECT_PATH, 'views/frontend/node-styles.css'),
         os.path.join(destination, 'node-styles.css')),
        (os.path.join(TEST_PROJECT_PATH, 'views/frontend/view-styles.css'),
         os.path.join(destination, 'view-styles.css')),
    ]

    for original_path, copy_path in static_files:

        # check existence
        assert os.path.exists(copy_path)

        # check content
        original = open(original_path, 'r').read()
        copy = open(copy_path, 'r').read()

        assert copy == original


# tests #######################################################################
def test_basic_collect_static():
    from lona.command_line.collect_static import collect_static

    with TemporaryDirectory() as tmp_dir:
        args = _generate_command_line_args(
            destination=tmp_dir,
        )

        collect_static(args=args)

        _check_javascript_client_src_files(destination=tmp_dir)
        _check_test_project_static_files(destination=tmp_dir)


def test_non_empty_destinations():
    from lona.command_line.collect_static import collect_static

    with TemporaryDirectory() as tmp_dir:

        # create a file so the destination is not empty
        test_file_path = os.path.join(tmp_dir, 'foo.txt')

        with open(test_file_path, 'w+') as f:
            f.write('foo')

        # first run
        args = _generate_command_line_args(
            destination=tmp_dir,
            clean=False,
        )

        collect_static(args=args)

        assert os.path.exists(test_file_path)
        _check_javascript_client_src_files(destination=tmp_dir)
        _check_test_project_static_files(destination=tmp_dir)

        # second run
        args = _generate_command_line_args(
            destination=tmp_dir,
            clean=True,
        )

        collect_static(args=args)

        assert not os.path.exists(test_file_path)
        _check_javascript_client_src_files(destination=tmp_dir)
        _check_test_project_static_files(destination=tmp_dir)


def test_overlapping_directories():
    """
    This test tests the handling of overlapping directories.

    If a project has two static directories, containing overlapping paths,
    the collect-static mechanism should override reoccurring paths.

    Example:

        Static Directory 1
            /directory/file1.txt          <- should be present in the output
            /directory/file2.txt          <- should be overridden in the output

        Static Directory 2
            /directory/file2.txt          <- should be present in the output
    """

    from lona.command_line.collect_static import collect_static

    static_dir1 = TemporaryDirectory()
    static_dir2 = TemporaryDirectory()
    destination = TemporaryDirectory()

    # setup tmp_dir1
    os.makedirs(os.path.join(static_dir1.name, 'test-directory'))

    path = os.path.join(static_dir1.name, 'test-directory/file1.txt')

    with open(path, 'w+') as f:
        f.write('tmp_dir1')

    path = os.path.join(static_dir1.name, 'test-directory/file2.txt')

    with open(path, 'w+') as f:
        f.write('tmp_dir1')

    # setup tmp_dir2
    os.makedirs(os.path.join(static_dir2.name, 'test-directory'))

    path = os.path.join(static_dir2.name, 'test-directory/file2.txt')

    with open(path, 'w+') as f:
        f.write('tmp_dir2')

    # run collect static
    args = _generate_command_line_args(
        static_dirs=[
            static_dir1.name,
            static_dir2.name,
        ],
        destination=destination.name,
        clean=False,
    )

    collect_static(args=args)

    # run checks
    file1 = os.path.join(destination.name, 'test-directory/file1.txt')
    file2 = os.path.join(destination.name, 'test-directory/file2.txt')

    assert os.path.exists(file1)
    assert os.path.exists(file2)

    file1_content = open(file1, 'r').read()
    file2_content = open(file2, 'r').read()

    assert file1_content == 'tmp_dir1'
    assert file2_content == 'tmp_dir1'



from playwright.async_api import async_playwright
import pytest

from lona.pytest import eventually
from lona.html import Select
from lona import View

GET_OPTIONS = 'e => Array.from(e.selectedOptions).map(option => option.value)'


@pytest.mark.parametrize('client_version', [1, 2])
@pytest.mark.parametrize('browser_name', ['chromium', 'firefox', 'webkit'])
async def test_selects(browser_name, client_version, lona_app_context):
    """
    This test tests HTML selects and multi selects, with and without pre
    selections, using a browser.
    """

    test_data = {}

    def setup_app(app):
        app.settings.CLIENT_VERSION = client_version

        # select ##############################################################
        @app.route('/select/nothing-selected/')
        class NothingSelected(View):
            def handle_request(self, request):
                select = Select(
                    values=[
                        ('foo', 'Foo'),
                        ('bar', 'Bar'),
                        ('baz', 'Baz'),
                    ],
                    bubble_up=True,
                )

                test_data['select/nothing-selected'] = select.value

                self.show(select)

                self.await_change(select)
                test_data['select/nothing-selected'] = select.value

        @app.route('/select/pre-selected/')
        class PreSelected(View):
            def handle_request(self, request):
                select = Select(
                    values=[
                        ('foo', 'Foo'),
                        ('bar', 'Bar', True),
                        ('baz', 'Baz'),
                    ],
                    bubble_up=True,
                )

                test_data['select/pre-selected'] = select.value

                self.show(select)

                self.await_change(select)
                test_data['select/pre-selected'] = select.value

        @app.route('/select/value-types/')
        class ValueTypes(View):
            def handle_request(self, request):
                select = Select(
                    values=[
                        (1,   'Integer'),
                        (1.0, 'Float'),
                        ('1', 'String'),
                    ],
                    bubble_up=True,
                )

                test_data['select/value-types'] = select.value

                self.show(select)

                self.await_change(select)
                test_data['select/value-types'] = select.value

        # multi select ########################################################
        @app.route('/multi-select/nothing-selected/')
        class MultiSelectNothingSelected(View):
            def handle_request(self, request):
                select = Select(
                    values=[
                        ('foo', 'Foo'),
                        ('bar', 'Bar'),
                        ('baz', 'Baz'),
                    ],
                    multiple=True,
                    bubble_up=True,
                )

                test_data['multi-select/nothing-selected'] = select.value

                self.show(select)

                self.await_change(select)
                test_data['multi-select/nothing-selected'] = select.value

        @app.route('/multi-select/pre-selected/')
        class MultiSelectPreSelected(View):
            def handle_request(self, request):
                select = Select(
                    values=[
                        ('foo', 'Foo', True),
                        ('bar', 'Bar', True),
                        ('baz', 'Baz'),
                    ],
                    multiple=True,
                    bubble_up=True,
                )

                test_data['multi-select/pre-selected'] = select.value

                self.show(select)

                self.await_change(select)
                test_data['multi-select/pre-selected'] = select.value

    context = await lona_app_context(setup_app)

    async with async_playwright() as p:
        browser = await getattr(p, browser_name).launch()
        browser_context = await browser.new_context()
        page = await browser_context.new_page()

        # select / nothing selected ###########################################
        # initial value
        await page.goto(context.make_url('/select/nothing-selected/'))
        await page.wait_for_selector('select')

        selected_options = await page.eval_on_selector('select', GET_OPTIONS)

        assert selected_options == ['foo']

        for attempt in eventually():
            async with attempt:
                assert test_data['select/nothing-selected'] == 'foo'

        # user select
        await page.select_option('select', 'bar')

        selected_options = await page.eval_on_selector('select', GET_OPTIONS)

        assert selected_options == ['bar']

        for attempt in eventually():
            async with attempt:
                assert test_data['select/nothing-selected'] == 'bar'

        # select / pre selected ###############################################
        # initial value
        await page.goto(context.make_url('/select/pre-selected/'))
        await page.wait_for_selector('select')

        selected_options = await page.eval_on_selector('select', GET_OPTIONS)

        assert selected_options == ['bar']

        for attempt in eventually():
            async with attempt:
                assert test_data['select/pre-selected'] == 'bar'

        # user select
        await page.select_option('select', 'foo')

        selected_options = await page.eval_on_selector('select', GET_OPTIONS)

        assert selected_options == ['foo']

        for attempt in eventually():
            async with attempt:
                assert test_data['select/pre-selected'] == 'foo'

        # select / value types ################################################
        # integer
        await page.goto(context.make_url('/select/value-types/'))
        await page.wait_for_selector('select')

        await page.select_option('select', label='Integer')
        selected_options = await page.eval_on_selector('select', GET_OPTIONS)

        assert selected_options == ['1']

        for attempt in eventually():
            async with attempt:
                assert test_data['select/value-types'] == '1'

        # float
        await page.goto(context.make_url('/select/value-types/'))
        await page.wait_for_selector('select')

        await page.select_option('select', label='Float')
        selected_options = await page.eval_on_selector('select', GET_OPTIONS)

        assert selected_options == ['1.0']

        for attempt in eventually():
            async with attempt:
                assert test_data['select/value-types'] == '1.0'

        # string
        await page.goto(context.make_url('/select/value-types/'))
        await page.wait_for_selector('select')

        await page.select_option('select', label='String')
        selected_options = await page.eval_on_selector('select', GET_OPTIONS)

        assert selected_options == ['1']

        for attempt in eventually():
            async with attempt:
                assert test_data['select/value-types'] == '1'

        # multi / nothing selected ############################################
        # initial value
        await page.goto(context.make_url('/multi-select/nothing-selected/'))
        await page.wait_for_selector('select')

        selected_options = await page.eval_on_selector('select', GET_OPTIONS)

        assert selected_options == []

        for attempt in eventually():
            async with attempt:
                assert test_data['multi-select/nothing-selected'] == []

        # user select
        await page.select_option('select', ['foo', 'bar'])

        selected_options = await page.eval_on_selector('select', GET_OPTIONS)

        assert selected_options == ['foo', 'bar']

        for attempt in eventually():
            async with attempt:
                assert test_data['multi-select/nothing-selected'] == [
                    'foo',
                    'bar',
                ]

        # multi / pre selected ################################################
        # initial value
        await page.goto(context.make_url('/multi-select/pre-selected/'))
        await page.wait_for_selector('select')

        selected_options = await page.eval_on_selector('select', GET_OPTIONS)

        assert selected_options == ['foo', 'bar']

        for attempt in eventually():
            async with attempt:
                assert test_data['multi-select/pre-selected'] == ['foo', 'bar']

        # user select
        await page.select_option('select', ['foo', 'baz'])

        selected_options = await page.eval_on_selector('select', GET_OPTIONS)

        assert selected_options == ['foo', 'baz']

        for attempt in eventually():
            async with attempt:
                assert test_data['multi-select/pre-selected'] == [
                    'foo',
                    'baz',
                ]

        # user deselect
        await page.goto(context.make_url('/multi-select/pre-selected/'))
        await page.wait_for_selector('select')
        await page.select_option('select', [])

        selected_options = await page.eval_on_selector('select', GET_OPTIONS)

        assert selected_options == []

        for attempt in eventually():
            async with attempt:
                assert test_data['multi-select/pre-selected'] == []



from lona.events.event_types import CHANGE
from lona.html.node import Node


class Option2(Node):
    TAG_NAME = 'option'

    def __init__(
            self,
            *args,
            value='',
            selected=False,
            disabled=False,
            render_value=True,
            **kwargs,
    ):

        super().__init__(*args, **kwargs)

        self.render_value = render_value
        self.value = value
        self.selected = selected
        self.disabled = disabled

    def _render_value(self, value):
        return str(value)

    # properties ##############################################################
    # value
    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, new_value):
        with self.lock:
            if self.render_value:
                self.attributes['value'] = self._render_value(new_value)

            self._value = new_value

    # selected
    @property
    def selected(self):
        return 'selected' in self.attributes

    @selected.setter
    def selected(self, new_value):
        if new_value:
            self.attributes['selected'] = ''

        else:
            del self.attributes['selected']

    # disabled
    @property
    def disabled(self):
        return 'disabled' in self.attributes

    @disabled.setter
    def disabled(self, new_value):
        if new_value:
            self.attributes['disabled'] = ''

        else:
            del self.attributes['disabled']


class Select2(Node):
    TAG_NAME = 'select'
    EVENTS = [CHANGE]

    def __init__(self, *options, disabled=False, multiple=False,
                 readonly=False, bubble_up=False, **kwargs):

        super().__init__(**kwargs)

        self.options = options
        self.disabled = disabled
        self.multiple = multiple
        self.readonly = readonly
        self.bubble_up = bubble_up

    def handle_input_event(self, input_event):
        if input_event.name != 'change':
            return super().handle_input_event(input_event)

        # select options by index
        selected_option_indexes = input_event.data

        with self.lock:
            for index, option in enumerate(self.options):
                option.selected = index in selected_option_indexes

        # run custom change event handler
        input_event = self.handle_change(input_event)

        if self.bubble_up:
            return input_event

    # select properties #######################################################
    # disabled
    @property
    def disabled(self):
        return 'disabled' in self.attributes

    @disabled.setter
    def disabled(self, new_value):
        if not isinstance(new_value, bool):
            raise TypeError('disabled is a boolean property')

        if new_value:
            self.attributes['disabled'] = ''

        else:
            del self.attributes['disabled']

    # multiple
    @property
    def multiple(self):
        return 'multiple' in self.attributes

    @multiple.setter
    def multiple(self, new_value):
        if not isinstance(new_value, bool):
            raise TypeError('multiple is a boolean property')

        if new_value:
            self.attributes['multiple'] = ''

        else:
            del self.attributes['multiple']

    # readonly
    @property
    def readonly(self):
        return 'readonly' in self.attributes

    @readonly.setter
    def readonly(self, new_value):
        if not isinstance(new_value, bool):
            raise TypeError('readonly is a boolean property')

        if new_value:
            self.attributes['readonly'] = ''

        else:
            del self.attributes['readonly']

    # option properties #######################################################
    # options
    @property
    def options(self):
        with self.lock:
            options = ()

            for node in self.nodes:
                if node.tag_name != 'option':
                    continue

                options += (node, )

            return options

    @options.setter
    def options(self, new_options):
        with self.lock:
            self.nodes.clear()

            if not isinstance(new_options, (list, tuple)):
                new_options = [new_options]

            for option in new_options:
                self.add_option(option)

    # selected options
    @property
    def selected_options(self):
        with self.lock:
            options = ()

            for option in self.options:
                if not option.selected:
                    continue

                options += (option, )

            if not options and self.options and not self.multiple:
                return (self.options[0], )

            return options

    # value
    @property
    def value(self):
        with self.lock:
            selected_options = self.selected_options
            values = ()

            for option in selected_options:
                values += (option.value, )

            if not self.multiple:
                if not values:
                    options = self.options

                    if options:
                        return self.options[0].value

                    return None

                return values[0]

            return values

    @value.setter
    def value(self, new_value):
        with self.lock:
            if not isinstance(new_value, list):
                new_value = [new_value]

            old_values = self.values

            for value in new_value:
                if value not in old_values:
                    raise RuntimeError(f'unknown value: {value}')

            for option in self.options:
                option.selected = option.value in new_value

    # values
    @property
    def values(self):
        with self.lock:
            values = ()

            for option in self.options:
                values += (option.value, )

            return values

    # helper ##################################################################
    def add_option(self, option):
        with self.lock:
            self.nodes.append(option)

    def remove_option(self, option):
        with self.lock:
            self.nodes.remove(option)

    def clear_options(self):
        with self.lock:
            for option in self.options:
                option.remove()

    def select_all(self):
        with self.lock:
            for option in self.options:
                option.selected = True

    def select_none(self):
        with self.lock:
            for option in self.options:
                option.selected = False



from lona.pytest import eventually
from lona import View


def setup_app(app):

    @app.route('/view-1/')
    class View1(View):
        def handle_request(self, request):
            if 'fire' in request.GET:
                self.fire_view_event('view1', {'view1': True})

        def on_view_event(self, view_event):
            self.server.state['view1'].append(
                (view_event.name, view_event.data),
            )

    @app.route('/view-2/')
    class View2(View):
        def handle_request(self, request):
            if 'fire' in request.GET:
                self.fire_view_event('view2', {'view2': True})

        def on_view_event(self, view_event):
            self.server.state['view2'].append(
                (view_event.name, view_event.data),
            )

    @app.route('/broadcast/')
    class FireBroadcastEvent(View):
        def handle_request(self, request):
            self.server.fire_view_event('broadcast', {'broadcast': True})


async def test_view_events(lona_app_context):
    """
    This test tests view events by creating two views that send events to all
    of their instances, and one view that sends events to all views in the
    system.
    The test is successful if both of the first two views only received their
    own events, and the broadcast events, and no event from view 1 leaked to
    view 2 or vice versa.
    """

    from playwright.async_api import async_playwright

    context = await lona_app_context(setup_app)

    async with async_playwright() as playwright:
        context.server.state['view1'] = []
        context.server.state['view2'] = []

        browser = await playwright.chromium.launch()
        browser_context = await browser.new_context()

        # view 1
        view1_page1 = await browser_context.new_page()
        view1_page2 = await browser_context.new_page()
        view1_page3 = await browser_context.new_page()

        await view1_page1.goto(context.make_url('/view-1/'))
        await view1_page2.goto(context.make_url('/view-1/'))
        await view1_page3.goto(context.make_url('/view-1/?fire'))

        # view 2
        view2_page1 = await browser_context.new_page()
        view2_page2 = await browser_context.new_page()
        view2_page3 = await browser_context.new_page()

        await view2_page1.goto(context.make_url('/view-2/'))
        await view2_page2.goto(context.make_url('/view-2/'))
        await view2_page3.goto(context.make_url('/view-2/?fire'))

        # broadcast
        broadcast_page = await browser_context.new_page()

        await broadcast_page.goto(context.make_url('/broadcast/'))

        # check captured events
        for attempt in eventually():
            async with attempt:
                assert sorted(context.server.state['view1']) == [
                    ('broadcast', {'broadcast': True}),
                    ('broadcast', {'broadcast': True}),
                    ('broadcast', {'broadcast': True}),
                    ('view1', {'view1': True}),
                    ('view1', {'view1': True}),
                    ('view1', {'view1': True}),
                ]

                assert sorted(context.server.state['view2']) == [
                    ('broadcast', {'broadcast': True}),
                    ('broadcast', {'broadcast': True}),
                    ('broadcast', {'broadcast': True}),
                    ('view2', {'view2': True}),
                    ('view2', {'view2': True}),
                    ('view2', {'view2': True}),
                ]



import os.path

from playwright.async_api import async_playwright

from lona.static_files import StyleSheet, Script
from lona import View


def setup_app(app):
    some_existing_file = os.path.basename(__file__)

    @app.route('/')
    class MyView(View):
        STATIC_FILES = [
            StyleSheet(
                name='my-view-style',
                path=some_existing_file,
                url='view.css',
            ),
            Script(
                name='my-view-script',
                path=some_existing_file,
                url='view.js',
            ),
            Script(
                name='my-view-map',
                path=some_existing_file,
                url='view.map',
                link=False,
            ),
            StyleSheet(
                name='my-view-disabled',
                path=some_existing_file,
                url='view.disabled',
                enabled_by_default=False,
            ),
        ]

        def handle_request(self, request):
            return 'SUCCESS'


async def test_static_files_view(lona_app_context):
    context = await lona_app_context(setup_app)

    assert (await context.client.get('/static/view.css')).status == 200
    assert (await context.client.get('/static/view.js')).status == 200
    assert (await context.client.get('/static/view.map')).status == 200
    assert (await context.client.get('/static/view.disabled')).status == 404

    async with async_playwright() as p:
        browser = await p.chromium.launch()
        browser_context = await browser.new_context()
        page = await browser_context.new_page()

        await page.goto(context.make_url())
        await page.wait_for_selector('#lona:has-text("SUCCESS")')

        links = await page.query_selector_all('link')
        scripts = await page.query_selector_all('script')
        style_urls = [await x.get_attribute('href') for x in links]
        script_urls = [await x.get_attribute('src') for x in scripts]
        html = await page.inner_html('html')

        assert '/static/view.css' in style_urls
        assert '/static/view.js' in script_urls
        assert 'view.map' not in html
        assert 'view.disabled' not in html



async def test_test_project(lona_project_context):
    import os

    from playwright.async_api import async_playwright

    TEST_PROJECT_PATH = os.path.join(__file__, '../../test_project')

    context = await lona_project_context(
        project_root=TEST_PROJECT_PATH,
        settings=['settings.py'],
    )

    async with async_playwright() as p:
        browser = await p.chromium.launch()
        browser_context = await browser.new_context()
        page = await browser_context.new_page()

        await page.goto(context.make_url())
        await page.wait_for_selector('#lona>h2:has-text("View Types")')



import re

import pytest

from lona.html import (
    NumberInput,
    TextInput,
    TextArea,
    CheckBox,
    Select2,
    Option2,
    Submit,
    Select,
    Option,
    Button,
    Span,
    Node,
    HTML,
    Div,
    H1,
)
from lona.compat import set_use_future_node_classes


@pytest.mark.incremental()
class TestAttributeDict:
    def test_get_initial_values(self):
        d = Div(foo='foo', bar='bar')

        assert d.attributes['foo'] == 'foo'
        assert d.attributes['bar'] == 'bar'

    def test_set_attributes(self):
        d = Div()

        d.attributes = {
            'foo': 'foo',
            'bar': 'bar',
        }

        assert d.attributes == {
            'foo': 'foo',
            'bar': 'bar',
        }

    def test_value_cant_be_dict(self):
        with pytest.raises(
                ValueError,
                match="unsupported type: <class 'dict'>",
        ):
            Div(foo={})

    def test_cant_use_id_key(self):
        d = Div()

        with pytest.raises(
                RuntimeError,
                match=re.escape(
                    "Node.attributes['id'] is not supported. "
                    'Use Node.id_list instead.',
                ),
        ):
            d.attributes['id'] = 'foo'

    def test_empty_dict_is_false(self):
        d = Div()

        assert not bool(d.attributes)

    def test_non_empty_dict_is_true(self):
        d = Div(foo='foo')

        assert bool(d.attributes)

    def test_pop_existing_key(self):
        d = Div(foo='foo-val', bar='bar-val')
        assert 'foo' in d.attributes

        val = d.attributes.pop('foo')

        assert val == 'foo-val'
        assert 'foo' not in d.attributes

    def test_pop_unknown_key(self):
        d = Div(foo='foo-val', bar='bar-val')
        assert 'xxx' not in d.attributes

        with pytest.raises(KeyError, match='xxx'):
            d.attributes.pop('xxx')
        assert 'xxx' not in d.attributes

    def test_pop_unknown_key_with_default(self):
        d = Div(foo='foo-val', bar='bar-val')
        assert 'xxx' not in d.attributes

        val = d.attributes.pop('xxx', 'yyy')

        assert val == 'yyy'
        assert 'xxx' not in d.attributes

    def test_pop_existing_key_with_default(self):
        d = Div(foo='foo-val', bar='bar-val')
        assert 'bar' in d.attributes

        val = d.attributes.pop('bar', 'yyy')

        assert val == 'bar-val'
        assert 'bar' not in d.attributes

    def test_pop_expects_at_most_2_arguments(self):
        d = Div(foo='foo-val', bar='bar-val')

        with pytest.raises(
                TypeError,
                match='pop expected at most 2 arguments, got 3',
        ):
            d.attributes.pop('xxx', 'yyy', 'zzz')

    def test_del_existing_key(self):
        d = Div(foo='foo-val', bar='bar-val')
        assert 'foo' in d.attributes

        del d.attributes['foo']

        assert 'foo' not in d.attributes

    def test_del_unknown_key(self):
        d = Div(foo='foo-val', bar='bar-val')
        assert 'xxx' not in d.attributes

        del d.attributes['xxx']

        assert 'xxx' not in d.attributes


@pytest.mark.incremental()
class TestAttributeList:
    def test_default_id_list_is_empty(self):
        d = Div()

        assert d.id_list == []

    def test_initial_id_can_be_list(self):
        d = Div(_id=['foo', 'bar'])

        assert d.id_list == ['foo', 'bar']

    def test_initial_id_can_be_space_separated_str(self):
        d = Div(_id='foo bar')

        assert d.id_list == ['foo', 'bar']

    def test_initial_id_can_be_passed_via_kwargs(self):
        d = Div(**{'id': 'foo bar'})

        assert d.id_list == ['foo', 'bar']

    def test_default_class_list_is_empty(self):
        d = Div()

        assert d.class_list == []

    def test_initial_class_can_be_list(self):
        d = Div(_class=['foo', 'bar'])

        assert d.class_list == ['foo', 'bar']

    def test_initial_class_can_be_space_separated_str(self):
        d = Div(_class='foo bar')

        assert d.class_list == ['foo', 'bar']

    def test_initial_class_can_be_passed_via_kwargs(self):
        d = Div(**{'class': 'foo bar'})

        assert d.class_list == ['foo', 'bar']

    def test_set_id_list(self):
        d = Div()

        d.id_list = ['foo', 'bar']

        assert d.id_list == ['foo', 'bar']

    def test_set_class_list(self):
        d = Div()

        d.class_list = ['foo', 'bar']

        assert d.class_list == ['foo', 'bar']

    def test_initial_id_cant_be_dict(self):
        with pytest.raises(
                ValueError,
                match='id has to be string or list of strings',
        ):
            Div(_id={})

    def test_cant_set_dict(self):
        d = Div()

        with pytest.raises(
                ValueError,
                match="unsupported type: <class 'dict'>",
        ):
            d.id_list = {}

    def test_cant_set_list_with_dict(self):
        d = Div()

        with pytest.raises(
                ValueError,
                match="unsupported type: <class 'dict'>",
        ):
            d.id_list = [{}]

    def test_default_list_has_zero_len(self):
        d = Div()

        assert len(d.id_list) == 0

    def test_len_returns_number_of_elements(self):
        d = Div(_id='foo bar')

        assert len(d.id_list) == 2

    def test_non_equality(self):
        d = Div(_id='foo bar')

        assert d.id_list != []
        assert d.id_list != ['foo', 'bar', 'baz']

    def test_equality_ignores_order(self):
        d = Div(_id='foo bar')

        assert d.id_list == ['bar', 'foo']

    def test_equality_ignored_duplicates(self):
        d = Div(_id='foo bar')

        assert d.id_list == ['foo', 'bar', 'foo', 'bar']

    def test_in_keyword(self):
        d = Div(_id='foo bar')

        assert 'foo' in d.id_list
        assert 'bar' in d.id_list
        assert 'baz' not in d.id_list

    def test_non_empty_list_is_true(self):
        d = Div(_id='foo')

        assert bool(d.id_list)

    def test_empty_list_is_false(self):
        d = Div()

        assert not bool(d.id_list)

    def test_add_one_element(self):
        d = Div()

        d.id_list.add('foo')

        assert d.id_list == ['foo']

    def test_add_existing_element_does_nothing(self):
        d = Div()
        d.id_list.add('foo')

        d.id_list.add('foo')

        assert d.id_list == ['foo']

    def test_cant_add_dict(self):
        d = Div()

        with pytest.raises(
                ValueError,
                match="unsupported type: <class 'dict'>",
        ):
            d.id_list.add({})

    def test_extend(self):
        d = Div(_id='foo')

        d.id_list.extend(['bar', 'baz'])

        assert d.id_list == ['foo', 'bar', 'baz']

    def test_extend_ignores_duplicates(self):
        d = Div(_id='foo')

        d.id_list.extend(['foo', 'bar', 'baz'])

        assert d.id_list == ['foo', 'bar', 'baz']

    def test_remove_existing_element(self):
        d = Div(_id='foo bar')

        d.id_list.remove('foo')

        assert d.id_list == ['bar']

    def test_remove_unknown_element(self):
        d = Div(_id='bar')

        d.id_list.remove('foo')

        assert d.id_list == ['bar']

    def test_clear(self):
        d = Div(_id='foo bar')

        d.id_list.clear()

        assert d.id_list == []

    def test_clear_empty_list(self):
        d = Div()

        d.id_list.clear()

        assert d.id_list == []

    def test_toggle_existing_element(self):
        d = Div(_id='foo bar')

        d.id_list.toggle('foo')

        assert d.id_list == ['bar']

    def test_toggle_unknown_element(self):
        d = Div(_id='bar')

        d.id_list.toggle('foo')

        assert d.id_list == ['foo', 'bar']


@pytest.mark.incremental()
class TestHTMLSubnodes:
    def test_nodes_in_nodes(self):
        div1 = Div()

        div2 = Div(
            div1,
        )

        assert len(div2) == 1
        assert div2[0] is div1

    def test_text_nodes_in_nodes(self):
        div = Div('foo')

        assert len(div) == 1
        assert str(div[0]) == 'foo'

    def test_mixed_nodes(self):
        sub_node1 = Div()
        sub_node2 = Div()

        div = Div(
            sub_node1,
            'foo',
            sub_node2,
            'bar',
        )

        assert len(div) == 4
        assert div[0] is sub_node1
        assert str(div[1]) == 'foo'
        assert div[2] is sub_node2
        assert str(div[3]) == 'bar'

    def test_generator(self):
        div = Div(
            (Div() for i in range(10)),
        )

        assert len(div) == 10

    def test_list(self):
        div = Div(
            [Div() for i in range(10)],
        )

        assert len(div) == 10


@pytest.mark.incremental()
class TestHTMLFromStr:
    def test_empty_node(self):
        node = HTML('<div></div>')[0]

        assert node.tag_name == 'div'
        assert node.id_list == []
        assert node.class_list == []
        assert node.style == {}
        assert node.attributes == {}
        assert node.nodes == []

    def test_node_with_attributes(self):
        node = HTML("""
            <div id="foo" class="bar" style="color: black" foo="bar">
            </div>
        """)[0]

        assert node.tag_name == 'div'
        assert node.id_list == ['foo']
        assert node.class_list == ['bar']
        assert node.style == {'color': 'black'}
        assert node.attributes == {'foo': 'bar'}
        assert node.nodes == []

    def test_sub_nodes(self):
        node = HTML("""
            <div>
                <span></span>
                <div></div>
                <h1></h1>
            </div>
        """)[0]

        assert node.tag_name == 'div'
        assert len(node.nodes) == 3
        assert node.nodes[0].tag_name == 'span'
        assert node.nodes[1].tag_name == 'div'
        assert node.nodes[2].tag_name == 'h1'

    def test_multiple_ids(self):
        node = HTML('<div id="foo bar baz"></div>')[0]

        assert node.id_list == ['foo', 'bar', 'baz']

    def test_multiple_classes(self):
        node = HTML('<div class="foo bar baz"></div>')[0]

        assert node.class_list == ['foo', 'bar', 'baz']

    def test_multiple_styles(self):
        node = HTML('<div style="color: black; display: block"></div>')[0]

        assert node.style == {
            'color': 'black',
            'display': 'block',
        }

    def test_multiple_attributes(self):
        node = HTML('<div foo="bar" bar="baz"></div>')[0]

        assert node.attributes == {
            'foo': 'bar',
            'bar': 'baz',
        }

    def test_high_level_nodes(self):
        node = HTML('<button>Click me</button>')[0]

        assert type(node) is Button

    def test_boolean_property_without_value(self):
        node = HTML('<button disabled>Click me</button>')[0]

        assert node.disabled

    def test_boolean_property_with_value(self):
        node = HTML('<button disabled="true">Click me</button>')[0]

        assert node.disabled

    def test_missing_end_tag(self):
        with pytest.raises(
                ValueError,
                match='Invalid html: missing end tag </span>',
        ):
            HTML('<span>')

    def test_wrong_end_tag(self):
        with pytest.raises(
                ValueError,
                match='Invalid html: </div> expected, </span> received',
        ):
            HTML('<p><div>abc</span></p>')

    def test_end_tag_without_start_tag(self):
        with pytest.raises(
                ValueError,
                match='Invalid html: missing start tag for </span>',
        ):
            HTML('<div>abc</div></span>')

    def test_missing_start_tag(self):
        with pytest.raises(
                ValueError,
                match='Invalid html: missing start tag for </span>',
        ):
            HTML('</span>')

    def test_self_closing_tag_with_slash(self):
        img = HTML('<div><img src="abc"/></div>')[0].nodes[0]

        assert img.tag_name == 'img'
        assert img.self_closing_tag is True

    def test_self_closing_tag_without_slash(self):
        img = HTML('<div><img src="abc"></div>')[0].nodes[0]

        assert img.tag_name == 'img'
        assert img.self_closing_tag is True

    def test_non_self_closing_tag(self):
        div = HTML('<div></div>')[0]

        assert div.tag_name == 'div'
        assert div.self_closing_tag is False

    def test_non_self_closing_tag_with_slash(self):
        span = HTML('<div><span/></div>')[0].nodes[0]

        assert span.tag_name == 'span'
        assert span.self_closing_tag is True

    def test_default_input_type_is_text(self):
        node = HTML('<input value="abc"/>')[0]

        assert type(node) is TextInput
        assert node.value == 'abc'
        assert node.disabled is False

    def test_input_type_text(self):
        node = HTML('<input type="text" value="xyz" disabled/>')[0]

        assert type(node) is TextInput
        assert node.value == 'xyz'
        assert node.disabled is True

    def test_input_type_unknown(self):
        node = HTML('<input type="unknown"/>')[0]

        assert type(node) is Node

    @pytest.mark.parametrize(
        'tp',
        [
            'button',
            'color',
            'date',
            'datetime-local',
            'email',
            'file',
            'hidden',
            'image',
            'month',
            'password',
            'radio',
            'range',
            'reset',  # intentionally, see 575dcf635180 ("html: remove Reset node")  # NOQA: LN002
            'search',
            'tel',
            'time',
            'url',
            'week',
        ],
    )
    def test_not_implemented_input_types(self, tp):
        node = HTML(f'<input type="{tp}"/>')[0]

        assert type(node) is Node

    def test_input_type_number(self):
        node = HTML('<input type="number" value="123.5"/>')[0]

        assert type(node) is NumberInput
        assert node.value == 123.5

    def test_input_type_checkbox(self):
        node = HTML('<input type="checkbox"/>')[0]

        assert type(node) is CheckBox
        assert node.value is False

    def test_input_type_checkbox_checked(self):
        node = HTML('<input type="checkbox" checked/>')[0]

        assert type(node) is CheckBox
        assert node.value is True

    def test_input_type_submit(self):
        node = HTML('<input type="submit"/>')[0]

        assert type(node) is Submit

    def test_textarea(self):
        node = HTML('<textarea>abc</textarea>')[0]

        assert type(node) is TextArea
        assert node.value == 'abc'

    def test_textarea_with_self_closing_tag_inside(self):
        textarea = HTML('<textarea>abc<br/>xyz</textarea>')[0]

        assert textarea.value == 'abc<br/>xyz'

    def test_textarea_with_pair_tag_inside(self):
        textarea = HTML('<textarea>aaa <b>bbb</b> ccc</textarea>')[0]

        assert textarea.value == 'aaa <b>bbb</b> ccc'

    def test_select(self):
        node = HTML("""
            <select>
                <option value="1">a</option>
                <option value="2" selected>b</option>
            </select>
        """)[0]

        assert type(node) == Select
        assert type(node.nodes[0]) == Option
        assert node.value == '2'

    def test_select2(self):
        set_use_future_node_classes(True)

        try:
            node = HTML("""
                <select>
                    <option value="1">a</option>
                    <option value="2" selected>b</option>
                </select>
            """)[0]

            assert type(node) == Select2
            assert type(node.nodes[0]) == Option2
            assert node.value == '2'

        finally:
            set_use_future_node_classes(False)


@pytest.mark.incremental()
class TestNumberInput:
    def test_default_properties(self):
        node = NumberInput()

        assert node.value is None
        assert node.min is None
        assert node.max is None
        assert node.step is None

    def test_initial_value(self):
        node = NumberInput(value=12.3)

        assert node.value == 12.3

    def test_change_value(self):
        node = NumberInput()
        node.value = 12.3

        assert node.value == 12.3

    def test_parsing_no_attributes(self):
        node = HTML('<input type="number"/>')[0]

        assert node.value is None
        assert node.min is None
        assert node.max is None
        assert node.step is None

    def test_parsing_int_value(self):
        node = HTML('<input type="number" value="123"/>')[0]

        assert node.value == 123

    def test_parsing_float_value(self):
        node = HTML('<input type="number" value="12.3"/>')[0]

        assert node.value == 12.3

    def test_parsing_broken_step(self):
        node = HTML('<input type="number" value="123" step="abc"/>')[0]

        assert node.value == 123
        assert node.step is None

    def test_parsing_int_step(self):
        node = HTML('<input type="number" value="12.3" step="3"/>')[0]

        assert node.value == 12.3
        assert node.step == 3

    def test_parsing_float_step(self):
        node = HTML('<input type="number" value="12.3" step="0.1"/>')[0]

        assert node.value == 12.3
        assert node.step == 0.1

    def test_parsing_broken_value(self):
        node = HTML('<input type="number" value="abc" step="0.1"/>')[0]

        assert node.raw_value == 'abc'
        assert node.value is None

    def test_parsing_all_attributes(self):
        node = HTML(
            '<input type="number" value="12.3" min="15.3" max="20.5" step="0.2"/>',
        )[0]

        assert node.value == 12.3
        assert node.min == 15.3
        assert node.max == 20.5
        assert node.step == 0.2

    def test_attribute_escaping(self):
        node = Div(style='font-family: "Times New Roman"')

        assert node.style['font-family'] == '"Times New Roman"'
        assert node.style.to_sub_attribute_string() == 'font-family: &quot;Times New Roman&quot;'  # NOQA: E501

        node = HTML(str(node))[0]

        assert node.style['font-family'] == '"Times New Roman"'

    # selectors ###############################################################
    def test_unsupported_selector(self):
        with pytest.raises(ValueError, match='unsupported selector feature:*'):
            HTML().query_selector('div > div')

    def test_query_selector(self):
        # test selectors by tag name
        html = Div(
            HTML(),
            'foo',
            Div(
                H1(),
            ),
            Span(),
            Div(),
        )

        assert html.query_selector('h1') is html[2][0]
        assert html.query_selector('span') is html[3]

        # test selectors by id
        html = Div(
            HTML(),
            'foo',
            Div(
                Div(_id='foo'),
                Div(_id='bar'),
            ),
            Div(_id='baz'),
        )

        assert html.query_selector('#foo') is html[2][0]
        assert html.query_selector('#baz') is html[3]

        # test selectors by class
        html = Div(
            HTML(),
            'foo',
            Div(
                Div(_class='foo'),
                Div(_class='bar'),
            ),
            Div(_class='baz'),
        )

        assert html.query_selector('.foo') is html[2][0]
        assert html.query_selector('.baz') is html[3]

        # test selectors by attribute
        html = Div(
            HTML(),
            'foo',
            Div(
                Div(_type='foo'),
                Div(_type='bar'),
            ),
            Div(_type='baz'),
        )

        assert html.query_selector('[type=foo]') is html[2][0]
        assert html.query_selector('[type=baz]') is html[3]

    def test_query_selector_all(self):
        html = Div(
            Span(),
            Div(
                Span(),
            ),
            Span(),
        )

        nodes = html.query_selector_all('span')

        assert len(nodes) == 3
        assert html[0] in nodes
        assert html[1][0] in nodes
        assert html[2] in nodes

    def test_closest(self):
        html = Div(
            Div(
                Div(
                    Span(),
                ),
                _id='foo',
            ),
        )

        span = html[0][0][0]
        node = span.closest('div#foo')

        assert node is html[0]

    # state ###################################################################
    def test_state(self):
        div = Div()

        assert hasattr(div, 'state')
        assert hasattr(div.state, 'lock')

        assert div.state == {}

        div.state['foo'] = 'bar'

        assert div.state == {'foo': 'bar'}

        div.state.clear()

        assert div.state == {}

    def test_initial_state(self):
        div = Div(state={'foo': 'bar'})

        assert div.state == {'foo': 'bar'}

    # slices ##################################################################
    def test_slices(self):
        div1 = Div()
        div2 = Div()
        div3 = Div()
        div4 = Div()
        outer_div = Div(div1, div2, div3, div4)
        outer_div.nodes = outer_div[1:-1]

        assert div1 not in outer_div
        assert div2 in outer_div
        assert div3 in outer_div
        assert div4 not in outer_div

    # loop detection ##########################################################
    def test_node_uniqueness(self):
        """
        Nodes are unique and may be mounted in only one node tree, at only one
        location. To ensure this, unmounts nodes at their parent, if they have
        a parent, before mounting them.
        """

        # multiple mounts of the same node
        # the resulting node should have only one child
        outer_div = Div()
        inner_div = Div()

        outer_div.append(inner_div)
        outer_div.append(inner_div)
        outer_div.append(inner_div)

        assert inner_div in outer_div
        assert len(outer_div) == 1

        # unmounting
        # the given node is already mounted somewhere else in the tree, so it
        # has to be unmounted before mounting it somewhere else
        inner_div = Div()

        outer_div = Div(
            Div(),
            Div(inner_div),
        )

        outer_div[0].append(inner_div)

        assert inner_div in outer_div[0]
        assert inner_div not in outer_div[1]

    def test_node_loop_detection(self):

        # simple loop
        div = Div(Div())

        with pytest.raises(RuntimeError, match='loop detected'):
            div[0].append(div)

        # multi node loop
        div = Div(Div(Div(Div())))

        with pytest.raises(RuntimeError, match='loop detected'):
            div[0][0][0].append(div[0][0])



import pytest

from lona import RedirectResponse, View
from lona.html import Button


def get_setup_app_callback(response_format):
    def setup_app(app):
        class RedirectButton(Button):
            def handle_input_event(self, input_event):
                if response_format == 'response-dict':
                    return {
                        'redirect': '/',
                    }

                elif response_format == 'response-class':
                    return RedirectResponse('/')

        @app.route('/')
        class Index(View):
            def handle_request(self, request):
                return 'SUCCESS'

        @app.route('/redirect-from-handle-input-event-root/')
        class RedirectFromHandleInputEventRootView(View):
            def handle_request(self, request):
                return RedirectButton()

            def handle_input_event_root(self, input_event):
                if response_format == 'response-dict':
                    return {
                        'redirect': '/',
                    }

                elif response_format == 'response-class':
                    return RedirectResponse('/')

        @app.route('/redirect-from-handle-input-event/')
        class RedirectFromHandleInputEventView(View):
            def handle_request(self, request):
                return Button()

            def handle_input_event(self, input_event):
                if response_format == 'response-dict':
                    return {
                        'redirect': '/',
                    }

                elif response_format == 'response-class':
                    return RedirectResponse('/')

        @app.route('/redirect-from-button/')
        class RedirectFromButtonView(View):
            def handle_request(self, request):
                return RedirectButton()

        @app.route('/redirect-from-on-view-event/')
        class RedirectFromOnViewEvent(View):
            def handle_request(self, request):
                return 'REDIRECT FROM ON VIEW EVENT'

            def on_view_event(self, view_event):
                if response_format == 'response-dict':
                    return {
                        'redirect': '/',
                    }

                elif response_format == 'response-class':
                    return RedirectResponse('/')

    return setup_app


@pytest.mark.parametrize(
    'response_format', ['response-dict', 'response-class'],
)
async def test_redirects_from_event_handlers(response_format, lona_app_context):
    """
    This test tests redirects from

     - View.handle_input_event_root()
     - View.handle_input_event()
     - View.on_view_event()
    """

    from playwright.async_api import async_playwright

    context = await lona_app_context(get_setup_app_callback(response_format))

    async with async_playwright() as p:
        browser = await p.chromium.launch()
        browser_context = await browser.new_context()
        page = await browser_context.new_page()

        # test redirect from View.handle_input_event_root()
        await page.goto(
            context.make_url('/redirect-from-handle-input-event-root/'),
        )

        await page.wait_for_url('/redirect-from-handle-input-event-root/')

        await page.click('button')

        await page.wait_for_url('/')
        await page.wait_for_selector('#lona:has-text("SUCCESS")')

        # test redirect from View.handle_input_event()
        await page.goto(
            context.make_url('/redirect-from-handle-input-event/'),
        )

        await page.wait_for_url('/redirect-from-handle-input-event/')

        await page.click('button')

        await page.wait_for_url('/')
        await page.wait_for_selector('#lona:has-text("SUCCESS")')

        # test redirect from button
        await page.goto(
            context.make_url('/redirect-from-button/'),
        )

        await page.wait_for_url('/redirect-from-button/')

        await page.click('button')

        await page.wait_for_url('/')
        await page.wait_for_selector('#lona:has-text("SUCCESS")')

        # test redirect from View.on_view_event()
        await page.goto(
            context.make_url('/redirect-from-on-view-event/'),
        )

        await page.wait_for_url('/redirect-from-on-view-event/')
        await page.wait_for_selector('#lona:has-text("REDIRECT FROM ON VIEW EVENT")')

        context.server.fire_view_event('foo')

        await page.wait_for_url('/')
        await page.wait_for_selector('#lona:has-text("SUCCESS")')



from lona import NotFoundError, View


def setup_app(app):

    @app.error_404_view
    class Error404NotFoundView(View):
        def handle_request(self, request):
            return 'NOT FOUND'

    @app.route('/url/<name>')
    class View2(View):
        def handle_request(self, request):
            if request.match_info['name'] == 'foo':
                raise NotFoundError

            return request.match_info['name']


async def test_404_not_found_errors(lona_app_context):
    from playwright.async_api import async_playwright

    context = await lona_app_context(setup_app)

    async with async_playwright() as playwright:
        context.server.state['view1'] = []
        context.server.state['view2'] = []

        browser = await playwright.chromium.launch()
        browser_context = await browser.new_context()
        page = await browser_context.new_page()

        await page.goto(context.make_url('/foo'))
        await page.wait_for_selector('body:has-text("NOT FOUND")')

        await page.goto(context.make_url('/url/foo'))
        await page.wait_for_selector('#lona:has-text("NOT FOUND")')

        await page.goto(context.make_url('/url/bar'))
        await page.wait_for_selector('#lona:has-text("bar")')



from aiohttp.web import Response


def setup_app(app):
    @app.route('/', http_pass_through=True)
    async def on_view_event(request):
        return Response(
            text='<h1>HTTP PASS THROUGH</h1>',
            content_type='text/html',
        )


async def test_http_pass_through(lona_app_context):
    from playwright.async_api import async_playwright

    context = await lona_app_context(setup_app)

    async with async_playwright() as p:
        browser = await p.chromium.launch()
        browser_context = await browser.new_context()
        page = await browser_context.new_page()

        await page.goto(context.make_url())

        await page.wait_for_selector('h1:has-text("HTTP PASS THROUGH")')



from playwright.async_api import async_playwright

from lona.html import NumberInput, Select, HTML
from lona.pytest import eventually
from lona import View


async def test_number_inputs(lona_app_context):
    test_data = {}

    def setup_app(app):

        @app.route('/')
        class NumberInputView(View):
            def handle_number_input_change(self, input_event):
                number_input = input_event.node

                test_data['valid'] = number_input.valid
                test_data['value'] = number_input.value
                test_data['raw_value'] = number_input.raw_value

            def handle_request(self, request):
                kwargs = {}

                for arg_name in ('min', 'max', 'step'):
                    if arg_name in request.GET:
                        kwargs[arg_name] = int(request.GET[arg_name])

                return HTML(
                    NumberInput(
                        handle_change=self.handle_number_input_change,
                        **kwargs,
                    ),
                    Select(),
                )

    context = await lona_app_context(setup_app)

    async with async_playwright() as p:
        browser = await p.chromium.launch()
        browser_context = await browser.new_context()
        page = await browser_context.new_page()

        # general tests #######################################################
        await page.goto(context.make_url('/'))
        await page.wait_for_selector('input[type=number]')

        test_data.clear()
        await page.fill('input[type=number]', '10')

        for attempt in eventually():
            async with attempt:
                assert test_data['valid'] is True
                assert test_data['value'] == 10
                assert test_data['raw_value'] == '10'

        # min #################################################################
        await page.goto(context.make_url('/?min=10'))
        await page.wait_for_selector('input[type=number]')

        test_data.clear()
        await page.fill('input[type=number]', '10')

        for attempt in eventually():
            async with attempt:
                assert test_data['valid'] is True
                assert test_data['value'] == 10
                assert test_data['raw_value'] == '10'

        test_data.clear()
        await page.fill('input[type=number]', '3')

        for attempt in eventually():
            async with attempt:
                assert test_data['valid'] is False
                assert test_data['value'] == 3
                assert test_data['raw_value'] == '3'

        # max #################################################################
        await page.goto(context.make_url('/?max=10'))
        await page.wait_for_selector('input[type=number]')

        test_data.clear()
        await page.fill('input[type=number]', '3')

        for attempt in eventually():
            async with attempt:
                assert test_data['valid'] is True
                assert test_data['value'] == 3
                assert test_data['raw_value'] == '3'

        test_data.clear()
        await page.fill('input[type=number]', '11')

        for attempt in eventually():
            async with attempt:
                assert test_data['valid'] is False
                assert test_data['value'] == 11
                assert test_data['raw_value'] == '11'

        # step ################################################################
        await page.goto(context.make_url('/?step=5'))
        await page.wait_for_selector('input[type=number]')

        test_data.clear()
        await page.fill('input[type=number]', '5')

        for attempt in eventually():
            async with attempt:
                assert test_data['valid'] is True
                assert test_data['value'] == 5
                assert test_data['raw_value'] == '5'

        test_data.clear()
        await page.fill('input[type=number]', '10')

        for attempt in eventually():
            async with attempt:
                assert test_data['valid'] is True
                assert test_data['value'] == 10
                assert test_data['raw_value'] == '10'

        test_data.clear()
        await page.fill('input[type=number]', '11')

        for attempt in eventually():
            async with attempt:
                assert test_data['valid'] is False
                assert test_data['value'] == 11
                assert test_data['raw_value'] == '11'



import logging
import os

logger = logging.getLogger('lona')

CLIENT_VERSION_ENV_VAR_NAME = 'LONA_CLIENT_VERSION'
CLIENT_VERSION_ENV_VAR_VALUES = (1, 2)

USE_FUTURE_NODE_CLASSES_ENV_VAR_NAME = 'LONA_USE_FUTURE_NODE_CLASSES'
USE_FUTURE_NODE_CLASSES_ENV_VAR_VALUES = (True, False)


# client version ##############################################################
def get_client_version():
    # TODO: remove in 2.0

    client_version = int(os.environ.get(CLIENT_VERSION_ENV_VAR_NAME, '1'))

    if client_version not in CLIENT_VERSION_ENV_VAR_VALUES:
        logger.error(
            'invalid env variable: %s=%s',
            CLIENT_VERSION_ENV_VAR_NAME,
            client_version,
        )

        return 1

    return client_version


def set_client_version(version):
    # TODO: remove in 2.0

    if version not in CLIENT_VERSION_ENV_VAR_VALUES:
        raise RuntimeError(f'invalid client version number: {version}')

    version = str(version)

    os.environ[CLIENT_VERSION_ENV_VAR_NAME] = version

    logger.debug('client version is set to %s', version)


# future node classes #########################################################
def get_use_future_node_classes():
    # TODO: remove in 2.0

    raw_use_future_node_classes = os.environ.get(
        USE_FUTURE_NODE_CLASSES_ENV_VAR_NAME,
        'false',
    )

    use_future_node_classes = {
        'true': True,
        'false': False,
        '1': True,
        '0': False,
    }[raw_use_future_node_classes.strip().lower()]

    if use_future_node_classes not in USE_FUTURE_NODE_CLASSES_ENV_VAR_VALUES:
        logger.error(
            'invalid env variable: %s=%s',
            USE_FUTURE_NODE_CLASSES_ENV_VAR_NAME,
            raw_use_future_node_classes,
        )

        return False

    return use_future_node_classes


def set_use_future_node_classes(enabled):
    # TODO: remove in 2.0

    if enabled not in USE_FUTURE_NODE_CLASSES_ENV_VAR_VALUES:
        raise RuntimeError(f'invalid value for {USE_FUTURE_NODE_CLASSES_ENV_VAR_NAME}: {enabled}')

    enabled = str(enabled)

    os.environ[USE_FUTURE_NODE_CLASSES_ENV_VAR_NAME] = enabled

    logger.debug('%s set to %s', USE_FUTURE_NODE_CLASSES_ENV_VAR_NAME, enabled)



from lona.exceptions import UserAbort
from lona.pytest import eventually
from lona.html import Button
from lona import View


def setup_app(app):

    @app.route('/')
    class HomeView(View):
        def handle_request(self, request):
            return 'HOME'

    @app.route('/test-view/')
    class TestView(View):
        def handle_request(self, request):
            self.show(Button())
            self.await_click()

            self.server.state['hooks'].append('handle_request')

        def on_stop(self, reason):
            self.server.state['hooks'].append('on_stop')
            self.server.state['stop_reason'] = reason

        def on_cleanup(self):
            self.server.state['hooks'].append('on_cleanup')


async def test_view_stop_hooks(lona_app_context):
    """
    This test tests the view hooks on_stop and on_cleanup
    """

    from playwright.async_api import async_playwright

    context = await lona_app_context(setup_app)

    async with async_playwright() as playwright:

        browser = await playwright.chromium.launch()
        browser_context = await browser.new_context()

        # go to page and stop it by clicking the button
        page1 = await browser_context.new_page()

        context.server.state['hooks'] = []
        context.server.state['stop_reason'] = 'NOTSET'

        await page1.goto(context.make_url('/test-view/'))
        await page1.wait_for_selector('button')

        for attempt in eventually():
            async with attempt:
                assert sorted(context.server.state['hooks']) == []
                assert context.server.state['stop_reason'] == 'NOTSET'

        # trigger stop
        await page1.click('button')

        for attempt in eventually():
            async with attempt:
                assert sorted(context.server.state['hooks']) == sorted([
                    'handle_request',
                    'on_stop',
                ])

                assert context.server.state['stop_reason'] is None

        # trigger cleanup
        await page1.goto(context.make_url('/'))
        await page1.wait_for_selector('#lona:has-text("HOME")')

        for attempt in eventually():
            async with attempt:
                assert sorted(context.server.state['hooks']) == sorted([
                    'handle_request',
                    'on_stop',
                    'on_cleanup',
                ])

        # go to page and stop it by reloading the page
        page2 = await browser_context.new_page()

        context.server.state['hooks'] = []
        context.server.state['stop_reason'] = 'NOTSET'

        await page2.goto(context.make_url('/test-view/'))
        await page2.wait_for_selector('button')

        for attempt in eventually():
            async with attempt:
                assert list(context.server.state['hooks']) == []
                assert context.server.state['stop_reason'] == 'NOTSET'

        await page2.reload()
        await page2.wait_for_selector('button')

        for attempt in eventually():
            async with attempt:
                assert sorted(context.server.state['hooks']) == sorted([
                    'on_stop',
                    'on_cleanup',
                ])

                assert isinstance(
                    context.server.state['stop_reason'],
                    UserAbort,
                )



async def test_click_events(lona_app_context):
    from playwright.async_api import async_playwright

    state = {
        'html': None,
        'input-events': [],
    }

    def setup_app(app):
        from lona.html import CLICK, HTML, Div
        from lona import View

        class OuterDiv(Div):
            EVENTS = [CLICK]

            STYLE = {
                'border': '1px solid black',
                'display': 'inline-flex',
                'background-color': 'lightgrey',
                'flex-direction': 'column',
                'height': '10em',
                'cursor': 'pointer',
                'padding': '1em',
            }

        class InnerDiv(Div):
            EVENTS = [CLICK]

            STYLE = {
                'border': '1px solid black',
            }

        class NonClickableInnerDiv(InnerDiv):
            EVENTS = []

        @app.route('/')
        class ClickTestView(View):
            def handle_input_event(self, input_event):
                state['input-events'].append(input_event)

                self.show('SUCCESS')

            def handle_request(self, request):
                html = HTML(
                    OuterDiv(
                        'Clickable Outer Div 1',
                        InnerDiv(
                            'Clickable Inner Div 1',
                            id='inner-div-1',
                        ),
                        id='outer-div-1',
                    ),
                    OuterDiv(
                        'Clickable Outer Div 2',
                        NonClickableInnerDiv(
                            'Non Clickable Inner Div 2',
                            id='inner-div-2',
                        ),
                        id='outer-div-2',
                    ),
                )

                self.show(html)
                state['html'] = html

    context = await lona_app_context(setup_app)

    async with async_playwright() as p:
        browser = await p.chromium.launch()
        browser_context = await browser.new_context()
        page = await browser_context.new_page()

        # clickable outer div
        state['input-events'].clear()

        await page.goto(context.make_url())
        assert await page.inner_text('#lona') != 'SUCCESS'
        await page.click('#outer-div-1')
        await page.wait_for_selector('#lona:has-text("SUCCESS")')

        outer_div = state['html'].query_selector('#outer-div-1')
        input_events = state['input-events']

        assert len(input_events) == 1
        assert input_events[0].node is outer_div
        assert input_events[0].target_node is outer_div

        # clickable inner div in clickable div
        state['input-events'].clear()

        await page.goto(context.make_url())
        assert await page.inner_text('#lona') != 'SUCCESS'
        await page.click('#inner-div-1')
        await page.wait_for_selector('#lona:has-text("SUCCESS")')

        inner_div = state['html'].query_selector('#inner-div-1')
        input_events = state['input-events']

        assert len(input_events) == 1
        assert input_events[0].node is inner_div
        assert input_events[0].target_node is inner_div

        # non clickable inner div in clickable div
        state['input-events'].clear()

        await page.goto(context.make_url())
        assert await page.inner_text('#lona') != 'SUCCESS'
        await page.click('#inner-div-2')
        await page.wait_for_selector('#lona:has-text("SUCCESS")')

        outer_div = state['html'].query_selector('#outer-div-2')
        inner_div = state['html'].query_selector('#inner-div-2')
        input_events = state['input-events']

        assert len(input_events) == 1
        assert input_events[0].node is outer_div
        assert input_events[0].target_node is inner_div



def setup_app(app):
    from lona import View

    # root
    @app.route('/')
    class RootView(View):
        def handle_request(self, request):
            return 'ROOT'

    @app.route('/redirect-to-root/')
    class RedirectToRoot(View):
        def handle_request(self, request):
            return {
                'redirect': '/',
            }

    # absolute url
    @app.route('/absolute-url/')
    class AbsoluteUrlView(View):
        def handle_request(self, request):
            return 'ABSOLUTE-URL'

    @app.route('/redirect-to-absolute-url/')
    class RedirectToAbsoluteUrlView(View):
        def handle_request(self, request):
            return {
                'redirect': '/absolute-url/',
            }

    # relative urls
    @app.route('/redirect-to-relative-url/foo/')
    class RelativeUrlView(View):
        def handle_request(self, request):
            return 'relative-URL'

    @app.route('/redirect-to-relative-url/')
    class RelativeRedirectUrlView(View):
        def handle_request(self, request):
            return {
                'redirect': 'foo/',
            }

    @app.route('/redirect-to-root-relatively/')
    class RedirectToRootRelativeView(View):
        def handle_request(self, request):
            return {
                'redirect': '..',
            }

    # refresh
    refreshed = False

    @app.route('/refresh/')
    class RefreshView(View):
        def handle_request(self, request):
            nonlocal refreshed

            if not refreshed:
                refreshed = True

                return {
                    'redirect': '.',
                }

            else:
                return 'REFRESH'


async def test_redirects(lona_app_context):
    """
    This test tests redirects by creating multiple views and redirecting from
    one to another. Absolute and relative URLs are tested.
    """

    from playwright.async_api import async_playwright

    context = await lona_app_context(setup_app)

    async with async_playwright() as p:
        browser = await p.chromium.launch()
        browser_context = await browser.new_context()
        page = await browser_context.new_page()

        # test redirect to root
        await page.goto(context.make_url('/redirect-to-root/'))
        await page.wait_for_url('/')
        await page.wait_for_selector('#lona:has-text("ROOT")')

        # test redirect to absolute url
        await page.goto(context.make_url('/redirect-to-absolute-url/'))
        await page.wait_for_url('/absolute-url/')
        await page.wait_for_selector('#lona:has-text("ABSOLUTE-URL")')

        # relative url
        await page.goto(context.make_url('/redirect-to-relative-url/'))
        await page.wait_for_url('/redirect-to-relative-url/foo/')
        await page.wait_for_selector('#lona:has-text("RELATIVE-URL")')

        await page.goto(context.make_url('/redirect-to-root-relatively/'))
        await page.wait_for_url('/')
        await page.wait_for_selector('#lona:has-text("ROOT")')

        # test refresh
        await page.goto(context.make_url('/refresh/'))
        await page.wait_for_url('/refresh/')
        await page.wait_for_selector('#lona:has-text("REFRESH")')
